#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdarg.h>
#include <errno.h>
#include <libgen.h>
#include <ctype.h>

#include "adwin-macros.h"
#include "linux/adwin.h"
#include "libadwin.h"
#include "libadwin/adctl.h"
#include "libadwin/bits.h"
#include "libadwin/errno.h"
#include "libadwin/types.h"
#include "configfile.h"
#include "retry_count.h"
#include "libadwin-macros.h"

#ifdef WITH_ETHERNET
#  include <sys/un.h>
#  include <netinet/in.h>
#  include <arpa/inet.h>
#  include <sys/time.h>
#  define USE_FOR_LIBADWIN 1
// todo:
#  include "ADEthernet.h"       /* Prototypes for Ethernet functions (NET_*) */
#  include "ethernet.h"         /* Prototypes for Ethernet functions (NET_*) */
#endif
#ifdef WITH_LINK
#  include "transputer.h" 	/* Prototypes for Transputer funktions (T_*) */
#  include "dsp.h" 		/* Prototypes for DSP funktions (DSP_*) */
#endif
#ifdef WITH_USB
#  include "usb.h" 		/* Prototypes for USB funktions (USB_*) */
#endif
#include "data_convert.h"

#if defined(__FreeBSD__)
#include <sys/syslimits.h>
#endif

const int32_t g_Version = 2000015;

void Set_DeviceNo(int32_t device_no) {
	adwin_dev_t Dev;
	int fd;
	DeviceNumber dn;
	DeviceNumber NewDn;
	char server_name[INET_ADDRSTRLEN];
	
	// check if the given device number is valid
	if ((device_no < 1) || (device_no > 0xfff)) {
		adwin_errno = k_this_link_not_allowed;
		return;
	}

	// copy CURRENTDEVICE do dn
	memcpy(&dn, &CURRENTDEVICE, sizeof (DeviceNumber));
	// if CURRENTDEVICE was already device_no, the number did not change
	if (dn.DeviceNo == device_no) return;

	if ((fd = SETMUTEX (device_no, &Dev)) >= 0) {
		inet_ntop(AF_INET, (const void*)&Dev.dev_net_server.sin_addr, 
			  server_name, INET_ADDRSTRLEN); // should not fail...
			
		if (Dev.dev_type == DC_ETHERNET)
			if (dn.DeviceType == DC_ETHERNET)
				reConnect(server_name,
					  Dev.dev_net_host_password,
					  Dev.dev_net_host_link,
					  ntohs (Dev.dev_net_server.sin_port),
					  Dev.dev_net_timeout);
			else
				openConnect(server_name,
					    Dev.dev_net_host_password,
					    Dev.dev_net_host_link,
					    ntohs (Dev.dev_net_server.sin_port),
					    Dev.dev_net_timeout);
		else
			if (dn.DeviceType == DC_ETHERNET)
				closeConnect ();

		NewDn.DeviceNo = device_no;
		NewDn.DeviceType = Dev.dev_type;
		memcpy (&CURRENTDEVICE, &NewDn, sizeof (DeviceNumber));

		CLEARMUTEX (&Dev, fd);
		adwin_errno = 0;
		return;
	}
	
	if (adwin_errno == k_this_link_is_unknown)
		return;
	adwin_errno = k_unknown_device;
}

/*****************************************************************************/

void Boot(const char *filename) {
	long res;

	res = adwin_boot(filename, 100000, CURRENTDEVICE.DeviceNo);

	if ((res > 1000) && (res != 100000) && (res != 8000)) {
		adwin_boot(filename, res, CURRENTDEVICE.DeviceNo);
	}
}

long adwin_boot(const char *btl, long iboardsize, int id) {
	adwin_dev_t dev;
	int fd;
	long tmp = 1;
	struct stat s;
	char share[PATH_MAX];

	if (stat(btl, &s) < 0) {
		if (! getAdwindir(share)) {
			adwin_errno = -errno;
			if (adwin_errno == -ENOENT) adwin_errno = k_File_not_found;
			return 1;
		}

		strcat(share, "/share/btl/");
		strcat(share, basename((char *)btl));

		if (stat(share, &s) < 0) {
			adwin_errno = -errno;
			if (adwin_errno == -ENOENT) adwin_errno = k_File_not_found;
			return 1;
		} else {
			btl = share;
		}
	}

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return 1;


 	switch(dev.dev_type) {
#ifdef WITH_ETHERNET
	case DC_ETHERNET: 
	SELECT_PROTOCOL(tmp, Boot, &dev, fd, btl, iboardsize);
/* 			NET_Boot(&dev, fd, btl, iboardsize); */
/* 			tmp = (adwin_errno = NET_Get_Last_Error ()) ? -1 : 0; */
	break;
#endif
#ifdef WITH_USB
	case DC_USB:
	  tmp = USB_Boot(&dev, fd, btl, iboardsize);
	  break;
#endif
#ifdef WITH_LINK
	case DC_LINK:
	case DC_PCMCIA: {
	  /* Am BTL-Namen determinieren welcher Prozessor */
	  char *btl_copy = strdup(btl);
	  int i;

	  for (i = 0; i < (signed)strlen(btl_copy); i++) 
	    btl_copy[i] = tolower(btl_copy[i]); 
	  
	  if (strstr(btl_copy, "adwin9.btl") != NULL) { 
	    tmp = DSP_Boot(&dev, fd, btl, iboardsize); 
	  } else { 
	    tmp = T_Boot(&dev, fd, btl, iboardsize); 
	  } 
	  free(btl_copy);
	  break;
	}
 #endif 
	default:
	  return 1;
 	}

	CLEARMUTEX(&dev, fd);

	return tmp == -1 ? 1 : tmp;
}

/*****************************************************************************/
int32_t Test_Version (){
	return adwin_test_version(CURRENTDEVICE.DeviceNo);
}

int adwin_test_version (int id) {
	adwin_dev_t dev;
	int fd, tmp = 3;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Test_Version, &dev, fd);

	CLEARMUTEX(&dev, fd);

	return tmp;
}

/*****************************************************************************/

processor_type_t Processor_Type(){
	return adwin_processor_type(CURRENTDEVICE.DeviceNo);
}

processor_type_t adwin_processor_type(int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	processor_type_t dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Processor_Type, &dev, fd, &dest);

/* 	switch (dev.dev_type) { */
/* #ifdef WITH_LINK */
/* 	case DC_LINK: */
/* 		switch (dev.dev_processor) { */
/* 		case DP_T2:  */
/* 			ret = PROCESSOR_T2; */
/* 			break; */
/* 		case DP_T4:  */
/* 			ret = PROCESSOR_T2;  */
/* 			break; */
/* 		case DP_T8:  */
/* 			ret = PROCESSOR_T8; */
/* 			break; */
/* 		case DP_DSP: */
/* 			ret = PROCESSOR_T9;  */
/* 			break; // there is no Link-T10 */
/* 		} */
/* 		break; */
/* #endif */
/* #ifdef WITH_USB */
/* 	case DC_UDB: */
/* 		tmp = USB_Processor_Type(&dev, fd, &dest); */
/* 		break; */
/* #endif */
/* #ifdef WITH_ETHERNET */
/* 	case DC_ETHERNET: */
/* 		// TODO! */
/* 		tmp = NET_Processor_Type(&dev, fd, &ret); */
/* 		dest = Generic_Driver_Processor_Type(); */
/* 		adwin_errno = Generic_Driver_Get_Last_Error(); */
/* #endif */
/* 	} */


	CLEARMUTEX(&dev, fd);


	return dest;
	// maybe todo
/*	adwin_dev_t dev;
	char proc[255];
	int fd;

	sprintf(proc, "/proc/adwin/%d/0x%X", getpid(), id);
	if ((fd = open(proc, O_RDONLY | O_EXCL)) < 0) {
		errno = ENOENT;
		return -1;
	}

	ioctl(fd, ADIOGKDEVINFO, &dev);
	close(fd);

	switch (dev.dev_processor) {
	case DP_21160:
	case DP_DSP: return 9;
	case DP_T2:  return 2;
	case DP_T4:  return 4;
	case DP_T8:  return 8;
	default:     return 0;
	} */
}

/*****************************************************************************/

int32_t Workload(int32_t priority) {
	return adwin_workload(priority, CURRENTDEVICE.DeviceNo);
}

int adwin_workload(int priority, int id) {
	int err, value;
	value = __adwin_workload(priority, &err, id);
	
	return ( err == -1) ? 255 : value;
}

int __adwin_workload(int priority, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	int32_t dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

        SELECT_PROTOCOL(tmp, Workload, &dev, fd, priority, &dest);

	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/

int32_t Free_Mem (mem_spec_t mem_spec){
	return adwin_mem(mem_spec, CURRENTDEVICE.DeviceNo);
}

long adwin_mem(mem_spec_t mem_spec, int id) {
	int err, memsize;
	memsize = __adwin_mem(mem_spec, &err, id);
	return (err == -1 ? 255 : memsize);
}

long __adwin_mem(mem_spec_t mem_spec, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	int32_t dest = -1;
	processor_type_t processor;
	
	/* first, we need to determine the processor type */
	processor = adwin_processor_type(id);
	/* todo: if this fails, we have two timeouts here... */

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Free_Mem, &dev, fd, mem_spec, &dest);

	switch (mem_spec) {
	case MEM_PM_LOCAL:
		if ((processor == PROCESSOR_T9) || (processor == PROCESSOR_T10))
			dest *= 6;
		else 
			dest *= 4;
		break;
	case MEM_ALL:
	case MEM_EM_LOCAL:
	case MEM_DM_LOCAL:
	case MEM_DRAM_EXTERN:
		dest *= 4;
		break;
	default:
		adwin_errno = -EINVAL;
		dest = -1;
	}
	CLEARMUTEX(&dev, fd);
	
	if (err)
		*err = adwin_errno;
	
	return dest;
}

/*****************************************************************************/

void Load_Process(const char *Filename) {
	adwin_process_load(Filename, CURRENTDEVICE.DeviceNo);
}

int adwin_process_load(const char *filename, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Load_Process, &dev, fd, filename);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 1;
}

/*****************************************************************************/

void Start_Process (int32_t process_no){
	adwin_process_start(process_no, CURRENTDEVICE.DeviceNo);
}

int adwin_process_start(int process_no, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Start_Process, &dev, fd, process_no * 2 + 99);

	CLEARMUTEX(&dev, fd);
	
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

void Stop_Process(int32_t ProcessNo){
	adwin_process_stop(ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_process_stop(int processID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;
	
	SELECT_PROTOCOL(tmp, Stop_Process, &dev, fd, 99 + 2 * processID);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

void Clear_Process (int32_t ProcessNo){
	adwin_process_clear(ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_process_clear(int processID, int id) {
	adwin_dev_t dev;
	int fd, tmp = 1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Clear_Process, &dev, fd, processID);

	CLEARMUTEX(&dev, fd);
	return (tmp != -1) ? 0 : 1;
}

/*****************************************************************************/

int32_t Process_Status (int32_t ProcessNo) {
	return adwin_process_status (ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_process_status (int ProcessNo, int id) {
	return adwin_get_par(ProcessNo - 100, id);
}
/*****************************************************************************/

void Set_Processdelay(int32_t ProcessNo, int32_t Value){
	adwin_set_processdelay(ProcessNo, Value, CURRENTDEVICE.DeviceNo);
}

int adwin_set_processdelay(int process, int delay, int id) {
	return adwin_set_par(process - 90, delay, id);
}

/*****************************************************************************/

int32_t Get_Processdelay(int32_t ProcessNo){
	return adwin_get_processdelay(ProcessNo, CURRENTDEVICE.DeviceNo);
}

int adwin_get_processdelay(int process, int id) {
	return adwin_get_par(process - 90, id);
}

int __adwin_get_processdelay(int process, int *err, int id) {
	return __adwin_get_par(process - 90, err, id);
}

/*****************************************************************************/
void Set_Par(int32_t Index, int32_t Value){
	adwin_set_par(Index, Value, CURRENTDEVICE.DeviceNo);
}

int adwin_set_par(int parID, int value, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Set_Par, &dev, fd, parID + 1000, value);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

int32_t Get_Par(int32_t Index) {
	return adwin_get_par(Index, CURRENTDEVICE.DeviceNo);
}

int adwin_get_par(int parID, int id) {
	int err, value;
	value = __adwin_get_par(parID, &err, id);

	return ( err == -1) ? 255 : value;
}

/* Linux-only */
int __adwin_get_par(int parID, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	int32_t dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Get_Par, &dev, fd, parID + 1000, &dest);

	CLEARMUTEX(&dev, fd);
	
	if (err)
		*err = tmp;
	
	return dest;
}

/*****************************************************************************/
void Get_Par_Block (int32_t dest[], int start_index, int count){
	adwin_get_par_block (start_index, count, dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_par_block (int first, int count, int32_t dest[], int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Get_Par_All, &dev, fd, first, count, dest);

	CLEARMUTEX(&dev, fd);
	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

void Get_Par_All(int32_t dest[]) {
	adwin_get_par_all (dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_par_all(int32_t dest[], int id) {
	return adwin_get_par_block (1, 80, dest, id);
}

/*****************************************************************************/

void Set_FPar(int32_t index, float value){
	adwin_set_fpar(index, value, CURRENTDEVICE.DeviceNo);
}

int adwin_set_fpar(int index, float value, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Set_FPar, &dev, fd, index + 1100, value);
	
	CLEARMUTEX(&dev, fd);

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

float Get_FPar(int32_t index) {
	return adwin_get_fpar(index, CURRENTDEVICE.DeviceNo);
}

float adwin_get_fpar(int index, int id) {
	int err;
	float value = __adwin_get_fpar(index, &err, id);

	return (err == -1) ? 255.0 : value;
}

float  __adwin_get_fpar(int index, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = -1;
	float dest = -1;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Get_FPar, &dev, fd, index + 1100, &dest);

	CLEARMUTEX(&dev, fd);
	
	if (err)
		*err = tmp;
	
	return dest;
}

/*****************************************************************************/

void Get_FPar_Block (float dest[], int start_index, int count){
	adwin_get_fpar_block (start_index, count, dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_block (int start_index, int count, float dest[], int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;
	
	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;
	
	SELECT_PROTOCOL(tmp, Get_FPar_All, &dev, fd, start_index + 100, count, dest);

	CLEARMUTEX(&dev, fd);

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

void Get_FPar_All(float dest[]){
	adwin_get_fpar_all(dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_all (float dest[], int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Get_FPar_All, &dev, fd, 101, 80, dest);

	CLEARMUTEX(&dev, fd);

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

void Get_FPar_Block_Double(double dest[], int start_index, int count){
	adwin_get_fpar_block_double(start_index, count, dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_block_double (int start_index, int count, double dest[], int id) {
	int i, tmp = 255;

	float fbuf[80];

	tmp = adwin_get_fpar_block(start_index, count, fbuf, id);
	
	if (tmp != -1)
		for (i = 0; i < count - start_index + 1; i++)
			dest[i] = (double) fbuf[i];

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/
void Get_FPar_All_Double (double dest[]){
	adwin_get_fpar_all_double(dest, CURRENTDEVICE.DeviceNo);
}

int adwin_get_fpar_all_double(double dest[], int id) {
	int i, tmp = 255;

	float fbuf[80];

	tmp = adwin_get_fpar_block(1, 80, fbuf, id);

	if (tmp != -1)
		for (i = 0; i < 80; i++)
			dest[i] = (double) fbuf[i];

	return tmp == -1 ? 255 : 0;
}

/*****************************************************************************/

int32_t Data_Length(int32_t data_no) {
	return adwin_data_length(data_no, CURRENTDEVICE.DeviceNo);
}

int adwin_data_length (int data_no, int id) {
	adwin_dev_t dev;
        int fd, tmp = 0;
	
        if ((fd = SETMUTEX(id, &dev)) < 0)
                return tmp;

        SELECT_PROTOCOL(tmp, Data_Length, &dev, fd, data_no);

        CLEARMUTEX(&dev, fd);
        return (tmp > -1) ? tmp : 0;
}

int32_t _Set_Data(int32_t data_no, void *src, adwin_data_type_t adt, int32_t start_index, int32_t count, int device_no) {
	adwin_dev_t dev;
	int fd, res;
	adwin_system_data_type_t asdt;
	void *ptr;
	
	asdt = adwin_get_data_type(data_no, device_no);

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(device_no, &dev)) < 0) {
		return 255;
	}

	if (asdt == -1) {
		// adwin_get_data_type failed - it should have set
		// adwin_errno appropriately, so just exit
		CLEARMUTEX(&dev, fd);
		return 255;
	} else if ((asdt != ADWIN_SYSTEM_SHORT) 
	    && (asdt != ADWIN_SYSTEM_INTEGER) 
	    && (asdt != ADWIN_SYSTEM_LONG) 
	    && (asdt != ADWIN_SYSTEM_FLOAT)) {
		// adwin_get_data_type returned some funny value,
		// which we map to k_Data_is_to_small here.
		adwin_errno = k_Data_is_to_small;
		CLEARMUTEX(&dev, fd);
		return 255;
	}

	ptr = Set_Data_convert(adt, src, asdt, count);

	if (!ptr ) {
		adwin_errno = -errno;
		CLEARMUTEX(&dev, fd);
		return 255;
	}
	
	SELECT_PROTOCOL_DATA(Set, res, &dev, fd, data_no, start_index, count, ptr, asdt);

	CLEARMUTEX(&dev, fd);

	if (ptr != src) free(ptr);
	return res ? 255 : 0;
}


int _Get_Data(int data_no, void *dest, adwin_data_type_t adt, int32_t start_index, int32_t count, int device_no) {
	adwin_dev_t dev;
	int fd, res;
	adwin_system_data_type_t asdt;
	void *ptr;
	
	asdt = adwin_get_data_type(data_no, device_no);
	
	if ((fd = SETMUTEX_INIT_GEN_DRIVER(device_no, &dev)) < 0) {
		return 255;
	}

	if (asdt == -1) {
		// adwin_get_data_type failed - it should have set
		// adwin_errno appropriately, so just exit
		CLEARMUTEX(&dev, fd);
		return 255;
	} else if ((asdt != ADWIN_SYSTEM_SHORT) 
	    && (asdt != ADWIN_SYSTEM_INTEGER) 
	    && (asdt != ADWIN_SYSTEM_LONG) 
	    && (asdt != ADWIN_SYSTEM_FLOAT)) {
		// adwin_get_data_type returned some funny value,
		// which we map to k_Data_is_to_small here.
		adwin_errno = k_Data_is_to_small;
		CLEARMUTEX(&dev, fd);
		return 255;
	}

	ptr = Get_Data_prepare_convert(asdt, dest, adt, count);
	
	if (!ptr ) {
		adwin_errno = -errno;
		CLEARMUTEX(&dev, fd);
		return 255;
	}
	
 	SELECT_PROTOCOL_DATA(Get, res, &dev, fd, data_no, start_index, count, ptr, asdt);

	CLEARMUTEX(&dev, fd);

	Get_Data_do_convert(asdt, ptr, adt, dest, count);

	if (ptr != dest) free(ptr);
	return res ? 255 : 0;
}

/*****************************************************************************/

void SetData_Long(int data_no, int32_t src[], int32_t start_index, int32_t count){
	_Set_Data(data_no, src, ADWIN_LONG, start_index, count, CURRENTDEVICE.DeviceNo);
}

/*****************************************************************************/

void GetData_Long(int data_no, int32_t dest[], int32_t start_index, int32_t count) {
	_Get_Data(data_no, dest, ADWIN_LONG, start_index, count, CURRENTDEVICE.DeviceNo);
}

/*****************************************************************************/

void SetData_Float(int data_no, float src[], int32_t start_index, int32_t count){
	_Set_Data(data_no, src, ADWIN_FLOAT, start_index, count, CURRENTDEVICE.DeviceNo);
}

/*****************************************************************************/

void SetData_Double(int data_no, double src[], int32_t start_index, int32_t count){
	_Set_Data(data_no, src, ADWIN_DOUBLE, start_index, count, CURRENTDEVICE.DeviceNo);
}

/*****************************************************************************/

void GetData_Float(int data_no, float dest[], int32_t start_index, int32_t count){
	_Get_Data(data_no, dest, ADWIN_FLOAT, start_index, count, CURRENTDEVICE.DeviceNo);
}

/*****************************************************************************/

void GetData_Double(int data_no, double dest[], int32_t start_index, int32_t count){
	_Get_Data(data_no, dest, ADWIN_DOUBLE, start_index, count, CURRENTDEVICE.DeviceNo);
}

/*****************************************************************************/

void Data2File (const char *filename, int data_no, int32_t start_index, int32_t count, int mode){
	adwin_data2file(filename, data_no, start_index, count, mode, CURRENTDEVICE.DeviceNo);
}

int adwin_data2file(const char *filename, int data, int32_t first, int32_t cnt, int mode, int id) {
	int fd;
	int32_t nbytes;
	char *ptr = NULL;
	
	switch(adwin_get_data_type(data, id)) {
	case 2:
		errno = ENOTSUP;
		return -1;
	case 3:
	case 4:
		nbytes = cnt *4;
		ptr = (char *) malloc(nbytes);
		if (_Get_Data(data, ptr, ADWIN_LONG, first, cnt, id)) {
			free(ptr);
			return -1;
		}
		break;
	case 5:
		nbytes = cnt *4;
		ptr = (char *) malloc(nbytes); 
		if (_Get_Data(data, ptr, ADWIN_FLOAT, first, cnt, id)) {
			free(ptr);
			return -1;
		}
		break;
	default:
		return -1;
	}
	
	errno = 0;
	fd = open(filename, O_WRONLY | O_CREAT | (mode ? O_APPEND : O_TRUNC), 0644);
	if (fd < 0) {
		free(ptr);
		perror("open");
		return -1;
	}
	
	errno = 0;
	if (write(fd, ptr, nbytes) < nbytes) {
		perror ("write");
		free(ptr);
		return -1;
	}
	
	free(ptr);
	close(fd);
	
	return 0;
}

/*****************************************************************************/

int32_t Fifo_Empty (int fifo_no){
	return adwin_fifo_empty(fifo_no, CURRENTDEVICE.DeviceNo);
}

int adwin_fifo_empty(int fifo_no, int id) {
	int err, cnt;
	cnt = __adwin_fifo_empty(fifo_no, &err, id);
	
	return ((err == -1) ? 255 : cnt);
}

int __adwin_fifo_empty (int fifo_no, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;
	int dest = 0;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;
	
	SELECT_PROTOCOL(tmp, Fifo_Empty, &dev, fd, fifo_no, &dest);
 
	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/

int32_t Fifo_Full (int fifo_no){
	return adwin_fifo_full(fifo_no, CURRENTDEVICE.DeviceNo);
}

int adwin_fifo_full (int fifo_no, int id) {
	int cnt, err;
	cnt = __adwin_fifo_full(fifo_no, &err, id);
	return (err == -1) ? 255 : cnt;
}

// todo: error handling

int __adwin_fifo_full (int fifo_no, int *err, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;
	int dest = 0;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Fifo_Full, &dev, fd, fifo_no, &dest);

	CLEARMUTEX(&dev, fd);

	if (err)
		*err = tmp;

	return dest;
}

/*****************************************************************************/

void Fifo_Clear (int fifo_no){
	adwin_fifo_clear(fifo_no, CURRENTDEVICE.DeviceNo);
}

int adwin_fifo_clear(int fifo_no, int id) {
	adwin_dev_t dev;
	int fd, tmp = 255;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(id, &dev)) < 0)
		return tmp;

	SELECT_PROTOCOL(tmp, Fifo_Clear, &dev, fd, fifo_no);

	CLEARMUTEX(&dev, fd);

	return (tmp != -1) ? tmp : 255;
}

/*****************************************************************************/
int _Set_Fifo(int32_t fifo_no, void *src, adwin_data_type_t adt, int32_t count, int32_t device_no) {
	adwin_dev_t dev;
	int fd, res;
	adwin_system_data_type_t asdt;
	void *ptr;
	
	asdt = adwin_get_data_type(fifo_no, device_no);

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(device_no, &dev)) < 0) {
		return 255;
	}

	if (asdt == -1) {
		// adwin_get_data_type failed - it should have set
		// adwin_errno appropriately, so just exit
		CLEARMUTEX(&dev, fd);
		return 255;
	} else if ((asdt != ADWIN_SYSTEM_SHORT) 
	    && (asdt != ADWIN_SYSTEM_INTEGER) 
	    && (asdt != ADWIN_SYSTEM_LONG) 
	    && (asdt != ADWIN_SYSTEM_FLOAT)) {
		// adwin_get_data_type returned some funny value,
		// which we map to k_Data_is_to_small here.
		adwin_errno = k_Data_is_to_small;
		CLEARMUTEX(&dev, fd);
		return 255;
	}

	ptr = Set_Data_convert(adt, src, asdt, count);

	if (!ptr ) {
		adwin_errno = -errno;
		CLEARMUTEX(&dev, fd);
		return 255;
	}
	SELECT_PROTOCOL_FIFO(Set, res, &dev, fd, fifo_no, count, ptr, asdt);

	CLEARMUTEX(&dev, fd);

	if (ptr != src) free(ptr);
	return res ? 255 : 0;
}


int _Get_Fifo(int fifo_no, void *dest, adwin_data_type_t adt, int32_t count, int device_no) {
	adwin_dev_t dev;
	int fd, res;
	adwin_system_data_type_t asdt;
	void *ptr;
	

	asdt = adwin_get_data_type(fifo_no, device_no);

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(device_no, &dev)) < 0) {
		return 255;
	}

	if (asdt == -1) {
		// adwin_get_data_type failed - it should have set
		// adwin_errno appropriately, so just exit
		CLEARMUTEX(&dev, fd);
		return 255;
	} else if ((asdt != ADWIN_SYSTEM_SHORT) 
	    && (asdt != ADWIN_SYSTEM_INTEGER) 
	    && (asdt != ADWIN_SYSTEM_LONG) 
	    && (asdt != ADWIN_SYSTEM_FLOAT)) {
		// adwin_get_data_type returned some funny value,
		// which we map to k_Data_is_to_small here.
		adwin_errno = k_Data_is_to_small;
		CLEARMUTEX(&dev, fd);
		return 255;
	}

	ptr = Get_Data_prepare_convert(asdt, dest, adt, count);
	
	if (!ptr ) {
		adwin_errno = -errno;
		CLEARMUTEX(&dev, fd);
		return 255;
	}

	SELECT_PROTOCOL_FIFO(Get, res, &dev, fd, fifo_no, count, ptr, asdt);

	CLEARMUTEX(&dev, fd);

	Get_Data_do_convert(asdt, ptr, adt, dest, count);

	if (ptr != dest) free(ptr);
	return res ? 255 : 0;
}

/*****************************************************************************/
void SetFifo_Long(int fifo_no, int32_t src[], int32_t count) {
	_Set_Fifo(fifo_no, src, ADWIN_LONG, count, CURRENTDEVICE.DeviceNo);
}
/*****************************************************************************/
void GetFifo_Long(int fifo_no, int32_t dest[], int32_t count){
	_Get_Fifo(fifo_no, dest, ADWIN_LONG, count, CURRENTDEVICE.DeviceNo);
}
/*****************************************************************************/
void SetFifo_Float (int fifo_no, float src[], int32_t count) {
	_Set_Fifo(fifo_no, src, ADWIN_FLOAT, count, CURRENTDEVICE.DeviceNo);
}
/*****************************************************************************/
void GetFifo_Float (int fifo_no, float dest[], int32_t count){
	_Get_Fifo(fifo_no, dest, ADWIN_FLOAT, count, CURRENTDEVICE.DeviceNo);
}
/*****************************************************************************/
void SetFifo_Double(int fifo_no, double src[], int32_t count){
	_Set_Fifo(fifo_no, src, ADWIN_DOUBLE, count, CURRENTDEVICE.DeviceNo);
}
/*****************************************************************************/
void GetFifo_Double (int fifo_no, double dest[], int32_t count){
	_Get_Fifo(fifo_no, dest, ADWIN_DOUBLE, count, CURRENTDEVICE.DeviceNo);
}
/*****************************************************************************/
void SetData_String (int data_no, char *src) {
	adwin_set_data_string(data_no, src, CURRENTDEVICE.DeviceNo);
}

int adwin_set_data_string(int data_no, const char *src, int id) {
	int32_t *ptr;
	int32_t tmp;
	size_t len, i;

	len = strlen(src);
	ptr = (int32_t *) malloc((len + 2) * sizeof(int32_t));

	ptr[0] = len;
	for (i = 0; i < len + 1; i++)
		ptr[i + 1] = (unsigned char)src[i];
	// just to be sure:
	ptr[len + 1] = '\0';
	
	tmp = _Set_Data(data_no, ptr, ADWIN_LONG, 1, len + 2, id);

	free(ptr);

	return (tmp != 0) ? -1 : len;
}
/*****************************************************************************/

int32_t GetData_String (int32_t data_no, char *dest, int32_t max_count){
	return adwin_get_data_string(data_no, max_count, dest, CURRENTDEVICE.DeviceNo);
}

int32_t adwin_get_data_string(int32_t data_no, int32_t count, char *dest, int32_t id) {
	int32_t i, len, *ptr;

	ptr = (int32_t *)malloc((count + 2) * sizeof(int32_t));

	// todo: malloc failure

	if (_Get_Data(data_no, ptr, ADWIN_LONG, 1, count + 1, id) != 0) {
		free(ptr);
		return -1;
	}
	
	len = ptr[0];
	if (count < len)
		len = count;
	
	for (i = 0; i < len; i++)
		dest[i] = (char) ptr[i + 1];
	
	dest[len] = '\0';
	
	free(ptr);
	return len;
}

/*****************************************************************************/
int32_t String_Length (int32_t data_no){
	return adwin_strlen(data_no, CURRENTDEVICE.DeviceNo);
}

int32_t adwin_strlen(int data_no, int id) {
	int32_t len;
	int ret;

	//ret = adwin_get_data_int32(data, 1, 1, &len, id);
	ret = _Get_Data(data_no, &len, ADWIN_LONG, 1, 1, id);
	// if an error occurred, return -1, otherwise length
	return ret ? -1 : len;
}
/*****************************************************************************/
int32_t Get_Last_Error () {
	int32_t temp = adwin_errno;
	adwin_errno = 0;
	return temp;
}

char *Get_Last_Error_Text (int32_t Errno) {
	return (char*)(adwin_strerror(Errno));
}
/*****************************************************************************/

int16_t Get_Known_DeviceNo(int32_t devices[], int32_t *count_devices) {
	int i;
#ifdef KERNEL_MODULE
	int conf_devices;
	struct adconf adc;
#endif 
	PRINTDEBUG(DBG_LIBADWIN, 
		   ("Get_Known_DeviceNo called with %p %d\n", devices, *count_devices));
#ifdef KERNEL_MODULE

	// get number of configured devices
	adctl(ADIOGDEVNUM, &conf_devices);

	// check if the given array is large enough
	if (*count_devices < conf_devices) {
		// if not, return required size
		*count_devices = conf_devices;
		return 1;
	}
	
	// determine necessary memory for the list of configured devices
	adc.adc_len = 0;
	adctl(ADIOGDEVCONF, &adc);
	if (! adc.adc_len) {
		// list is empty -- let the caller know we did not
		// copy anything and return
		*count_devices = conf_devices;
		return 0;
	}

	// allocate memory for list
	adc.adc_buf = malloc(adc.adc_len);
	// get the list of configured devices
        adctl(ADIOGDEVCONF, &adc);
	
	// copy device numbers to array
	for (i = 0; i < conf_devices; i++) {
		devices[i] = strtol(adc.adc_req[i].adr_name, NULL, 0);
	}
	// free the local list
	free(adc.adc_buf);

	// let the caller know how much we copied
	*count_devices = conf_devices;
#else
	// check if the given array is large enough
	if (*count_devices < getDeviceCount()) {
		// if not, return required size
		*count_devices = getDeviceCount();
		return 1;
	}

	// copy device numbers to array
	for (i = 0; i < getDeviceCount(); i++) {
		adwin_dev_t tmp = getDevice(i);
		devices[i] = (short)tmp.dev_name;
	}

	// let the caller know how much we copied
	*count_devices = getDeviceCount();
#endif	
	return 0;
}

/*****************************************************************************/

const char* Get_Library_Version() {
	return VersionString;
}

char *Get_Password() {
	adwin_dev_t dev;
	char *buf = calloc(1, PWLEN + 1);
	int fd;

	if ((fd = SETMUTEX_INIT_GEN_DRIVER(CURRENTDEVICE.DeviceNo, &dev)) < 0)
		return NULL;

	CLEARMUTEX(&dev, fd);
	strncpy(buf, dev.dev_net_host_password, PWLEN);
	
	return buf;
}
